from abc import ABCMeta, abstractmethod


class InputInterceptionDataHandler(object):
    """
    A class that act as a pluggable hook that can be used during input interception when recording and playing the data
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def prepare_input_for_recording(self, interception_key, result, args, kwargs):
        """
        Prepare the input result that should be saved in the recording
        :param interception_key: Input interception key
        :type interception_key: basestring
        :param result: Input function invocation result (the input)
        :type result: Any
        :param args: Input invocation args
        :type args: tuple
        :param kwargs: Input invocation kwargs
        :type kwargs: dict
        :return: Input result in a form that should be saved in the recording
        :rtype: Any
        """
        pass

    @abstractmethod
    def restore_input_from_recording(self, recorded_data, args, kwargs):
        """
        Restore the actual input from the recording
        :param recorded_data: Recorded data provided by the prepare method
        :type recorded_data: Any
        :param args: Input invocation args
        :type args: tuple
        :param kwargs: Input invocation kwargs
        :type kwargs: dict
        :return: Input saved to the recording
        :rtype: Any
        """
        pass



import uuid
from abc import ABCMeta, abstractmethod


class Recording(object):
    """
    Holds a recording of an operation that was recorded using the TapeRecorder
    """

    __metaclass__ = ABCMeta

    def __init__(self, _id=None):
        self.id = _id or uuid.uuid1().hex
        self._closed = False

    @abstractmethod
    def _set_data(self, key, value):
        """
        Sets data in the recording
        :param key: data key
        :type key: basestring
        :param value: data value (serializable)
        :type value: Any
        """
        pass

    def set_data(self, key, value):
        """
        Sets data in the recording
        :param key: data key
        :type key: basestring
        :param value: data value (serializable)
        :type value: Any
        """
        assert not self._closed
        self._set_data(key, value)

    @abstractmethod
    def get_data(self, key):
        """
        :param key: Data key
        :type key: basestring
        :return: Recorded data under given key
        :rtype: Any
        :raise: playback.exceptions.RecordingKeyError
        """
        pass

    @abstractmethod
    def get_all_keys(self):
        """
        :return: All recorded keys
        :rtype: list of basestring
        """
        pass

    def add_metadata(self, metadata):
        """
        :param metadata: Metadata to add to the recording
        :type metadata: dict
        """
        assert not self._closed
        self._add_metadata(metadata)

    @abstractmethod
    def _add_metadata(self, metadata):
        """
        :param metadata: Metadata to add to the recording
        :type metadata: dict
        """
        pass

    def close(self):
        """
        Close the recording and release any underlying resources
        """
        self._closed = True

    def __setitem__(self, key, value):
        self._set_data(key, value)

    def __getitem__(self, item):
        return self.get_data(item)

    @abstractmethod
    def get_metadata(self):
        """
        :return: Recorded metadata
        :rtype: dict
        """
        pass



from playback.interception.files.file_interception import FileInterception
from playback.interception.input_interception import InputInterceptionDataHandler


class InputInterceptionFileDataHandler(InputInterceptionDataHandler, FileInterception):
    """
    Intercept file arguments for playback
    """
    def prepare_input_for_recording(self, interception_key, result, args, kwargs):
        """
        Reads the intercepted file, encode its content as string and return it ready for recording
        :param interception_key: Input interception key
        :type interception_key: basestring
        :param result: Input function invocation result (the input)
        :type result: Any
        :param args: Input invocation args
        :type args: tuple
        :param kwargs: Input invocation kwargs
        :type kwargs: dict
        :return: Input result in a form that should be saved in the recording
        :rtype: dict[str, str]
        """
        return self._intercept_file(args, kwargs)

    def restore_input_from_recording(self, recorded_data, args, kwargs):
        """
        Create a file from the recorded content and place it in the given file path
        :param recorded_data: Recorded data provided by the prepare method
        :type recorded_data: Any
        :param args: Input invocation args
        :type args: tuple
        :param kwargs: Input invocation kwargs
        :type kwargs: dict
        :return: Path of restored file
        :rtype: basestring
        """
        file_path = self._get_file_path(args, kwargs)
        with open(file_path, "wb") as binary_file:
            __, file_content = self._deserialize_file(recorded_data)
            binary_file.write(file_content)

        return file_path



# p3ready
from __future__ import absolute_import
from __future__ import print_function
import unittest
from random import shuffle, random

from jsonpickle import encode, decode
from mock import patch
from playback.exceptions import InputInterceptionKeyCreationError, RecordingKeyError
from playback.interception.input_interception import InputInterceptionDataHandler
from time import sleep

from playback.interception.output_interception import OutputInterceptionDataHandler
from playback.tape_recorder import TapeRecorder, CapturedArg, RecordingParameters, pickle_copy
from playback.tape_cassettes.in_memory.in_memory_tape_cassette import InMemoryTapeCassette
import six
from six.moves import range


class TestTapeRecorder(unittest.TestCase):

    def setUp(self):
        self.tape_cassette = InMemoryTapeCassette()
        self.tape_recorder = TapeRecorder(self.tape_cassette)
        self.tape_recorder.enable_recording()

    def test_record_and_playback_basic_operation_no_parameters(self):
        class Operation(object):

            @self.tape_recorder.operation()
            def execute(self):
                return 5

        instance = Operation()
        result = instance.execute()
        self.assertEqual(5, result)

        recording_id = self.tape_cassette.get_last_recording_id()
        playback_result = self.tape_recorder.play(recording_id,
                                                  playback_function=lambda recording: Operation().execute())
        self._assert_playback_vs_recording(playback_result, result)

    def _assert_playback_vs_recording(self, playback_result, result):
        """
        :param playback_result: Playback result
        :type playback_result: playback.tape_recorder.Playback
        :param result: Operation result
        :type result: Any
        """
        if six.PY3:
            self.assertCountEqual(playback_result.recorded_outputs, playback_result.playback_outputs)
        else:
            self.assertItemsEqual(playback_result.recorded_outputs, playback_result.playback_outputs)

        operation_output = next(po for po in playback_result.playback_outputs
                                if TapeRecorder.OPERATION_OUTPUT_ALIAS in po.key)
        self.assertEqual(result, operation_output.value['args'][0])
        self.assertGreater(playback_result.playback_duration, 0)
        self.assertGreater(playback_result.recorded_duration, 0)

    def test_record_and_playback_basic_operation_data_interception_no_arguments(self):
        class Operation(object):

            def __init__(self, seed=0):
                self.seed = seed

            @self.tape_recorder.operation()
            def execute(self):
                return self.get_value()

            @self.tape_recorder.intercept_input('input')
            def get_value(self):
                return self.seed

        instance = Operation(5)
        result = instance.execute()
        self.assertEqual(5, result)

        recording_id = self.tape_cassette.get_last_recording_id()
        playback_result = self.tape_recorder.play(recording_id,
                                                  playback_function=lambda recording: Operation().execute())
        self._assert_playback_vs_recording(playback_result, result)

    def test_record_and_playback_basic_operation_data_interception_property(self):
        class Operation(object):

            def __init__(self, seed=0):
                self.seed = seed

            @self.tape_recorder.operation()
            def execute(self):
                return self.get_value

            @self.tape_recorder.intercept_input('input')
            @property
            def get_value(self):
                return self.seed

        instance = Operation(5)
        result = instance.execute()
        self.assertEqual(5, result)

        recording_id = self.tape_cassette.get_last_recording_id()
        playback_result = self.tape_recorder.play(recording_id,
                                                  playback_function=lambda recording: Operation().execute())
        self._assert_playback_vs_recording(playback_result, result)

    def test_record_and_playback_basic_operation_data_interception_with_arguments(self):
        class Operation(object):

            def __init__(self, seed=0):
                self.seed = seed

            @self.tape_recorder.operation()
            def execute(self):
                val1 = self.get_value(2, b=3)
                val2 = self.get_value(4, b=6)
                return val1 + val2

            @self.tape_recorder.intercept_input('input')
            def get_value(self, a, b=2):
                return (a + b) * self.seed

        instance = Operation(seed=1)
        result = instance.execute()
        self.assertEqual(15, result)

        recording_id = self.tape_cassette.get_last_recording_id()
        playback_result = self.tape_recorder.play(recording_id,
                                                  playback_function=lambda recording: Operation().execute())
        self._assert_playback_vs_recording(playback_result, result)

    def test_record_and_playback_basic_operation_static_data_interception_with_arguments(self):
        class Operation(object):

            seed = 0

            @self.tape_recorder.operation()
            def execute(self):
                val1 = self.get_value(2, b=3)
                val2 = self.get_value(4, b=6)
                return val1 + val2

            @staticmethod
            @self.tape_recorder.static_intercept_input('input')
            def get_value(a, b=2):
                return (a + b) * Operation.seed

        instance = Operation()
        Operation.seed = 1
        result = instance.execute()
        self.assertEqual(15, result)

        Operation.seed = 0
        recording_id = self.tape_cassette.get_last_recording_id()
        playback_result = self.tape_recorder.play(recording_id,
                                                  playback_function=lambda recording: Operation().execute())
        self._assert_playback_vs_recording(playback_result, result)

    def test_record_and_playback_basic_operation_data_interception_with_arguments_and_metadata(self):
        def operation_metadata_extractor(obj):
            return {'type': type(obj)}

        class Operation(object):

            def __init__(self, seed=0):
                self.seed = seed

            @self.tape_recorder.operation(metadata_extractor=operation_metadata_extractor)
            def execute(self):
                val1 = self.get_value(2, b=3)
                val2 = self.get_value(4, b=6)
                return val1 + val2

            @self.tape_recorder.intercept_input('input')
            def get_value(self, a, b=2):
                return (a + b) * self.seed

        instance = Operation(seed=1)
        result = instance.execute()
        self.assertEqual(15, result)

        recording_id = self.tape_cassette.get_last_recording_id()
        recording = self.tape_cassette.get_recording(recording_id)
        # pickle encode decode doesn't reconstruct actuall class if this is an inner class
        operation_decoded = decode(encode(Operation, unpicklable=False))
        self.assertDictContainsSubset({'type': operation_decoded, TapeRecorder.OPERATION_CLASS: operation_decoded},
                                      recording.get_metadata())
        playback_result = self.tape_recorder.play(recording_id,
                                                  playback_function=lambda recording: Operation().execute())
        self._assert_playback_vs_recording(playback_result, result)

    def test_record_and_playback_basic_operation_no_parameters_with_output(self):
        class Operation(object):

            @self.tape_recorder.operation()
            def execute(self):
                x = 0
                x += self.output(4, arg='a')
                x += self.output(3, arg='b')
                return x

            @self.tape_recorder.intercept_output('output_function')
            def output(self, value, arg=None):
                return value

        instance = Operation()
        result = instance.execute()
        self.assertEqual(7, result)

        recording_id = self.tape_cassette.get_last_recording_id()
        playback_result = self.tape_recorder.play(recording_id,
                                                  playback_function=lambda recording: Operation().execute())

        self._assert_playback_vs_recording(playback_result, result)

        self.assertEqual({'args': [4], 'kwargs': {'arg': 'a'}}, playback_result.playback_outputs[0].value)
        self.assertEqual({'args': [3], 'kwargs': {'arg': 'b'}}, playback_result.playback_outputs[1].value)
        self.assertNotEqual(playback_result.playback_outputs[0].key, playback_result.playback_outputs[1].key)
        self.assertIn('output_function', playback_result.playback_outputs[0].key)
        self.assertIn('output_function', playback_result.playback_outputs[1].key)

    def test_record_and_playback_basic_operation_no_parameters_with_static_output(self):
        class Operation(object):

            @self.tape_recorder.operation()
            def execute(self):
                x = 0
                x += self.output(4, arg='a')
                x += self.output(3, arg='b')
                return x

            @staticmethod
            @self.tape_recorder.static_intercept_output('output_function')
            def output(value, arg=None):
                return value

        instance = Operation()
        result = instance.execute()
        self.assertEqual(7, result)

        recording_id = self.tape_cassette.get_last_recording_id()
        playback_result = self.tape_recorder.play(recording_id,
                                                  playback_function=lambda recording: Operation().execute())

        self._assert_playback_vs_recording(playback_result, result)

        self.assertEqual({'args': [4], 'kwargs': {'arg': 'a'}}, playback_result.playback_outputs[0].value)
        self.assertEqual({'args': [3], 'kwargs': {'arg': 'b'}}, playback_result.playback_outputs[1].value)
        self.assertNotEqual(playback_result.playback_outputs[0].key, playback_result.playback_outputs[1].key)
        self.assertIn('output_function', playback_result.playback_outputs[0].key)
        self.assertIn('output_function', playback_result.playback_outputs[1].key)

    def test_record_and_playback_basic_operation_similar_duration(self):
        class Operation(object):

            @self.tape_recorder.operation()
            def execute(self):
                sleep(0.5)
                return 5

        instance = Operation()
        result = instance.execute()
        self.assertEqual(5, result)

        recording_id = self.tape_cassette.get_last_recording_id()
        playback_result = self.tape_recorder.play(recording_id,
                                                  playback_function=lambda recording: Operation().execute())
        self.assertLessEqual(abs(playback_result.recorded_duration - playback_result.playback_duration), 0.1)

    def test_record_and_playback_basic_operation_class_extraction_in_metadata(self):
        class Operation(object):

            @self.tape_recorder.operation()
            def execute(self):
                return 5

            @classmethod
            @self.tape_recorder.class_operation()
            def class_execute(cls):
                return 5

        instance = Operation()
        self.assertEqual(5, instance.execute())

        recording_id = self.tape_cassette.get_last_recording_id()
        recording = self.tape_cassette.get_recording(recording_id)
        # pickle encode decode doesn't reconstruct actuall class if this is an inner class
        operation_decoded = decode(encode(Operation, unpicklable=False))
        self.assertDictContainsSubset({TapeRecorder.OPERATION_CLASS: operation_decoded}, recording.get_metadata())

        self.assertEqual(5, instance.class_execute())

        recording_id = self.tape_cassette.get_last_recording_id()
        recording = self.tape_cassette.get_recording(recording_id)
        # Inner classes are not unpicklable back to class upon decode
        self.assertDictContainsSubset({TapeRecorder.OPERATION_CLASS: operation_decoded}, recording.get_metadata())

    def test_drop_recording_direct_api_call(self):
        local_tape_recorder = self.tape_recorder

        class Operation(object):

            @self.tape_recorder.operation()
            def execute(self):
                local_tape_recorder.discard_recording()
                return 5

        instance = Operation()
        with patch.object(InMemoryTapeCassette, '_save_recording', wraps=self.tape_cassette._save_recording) \
                as intercepted_save, \
                patch.object(InMemoryTapeCassette, 'abort_recording', wraps=self.tape_cassette.abort_recording) \
                as intercepted_abort:
            result = instance.execute()
            intercepted_save.assert_not_called()
            intercepted_abort.assert_called()

        self.assertEqual(5, result)

    def test_skip_recording_decorator(self):

        @self.tape_recorder.recording_params(RecordingParameters(skipped=True))
        class Operation(object):

            @self.tape_recorder.operation()
            def execute(self):
                return 5

        instance = Operation()
        with patch.object(InMemoryTapeCassette, 'create_new_recording', wraps=self.tape_cassette.create_new_recording) \
                as intercepted:
            result = instance.execute()
            intercepted.assert_not_called()

        self.assertEqual(5, result)

    def test_skip_recording_decorator_compatible(self):
        @self.tape_recorder.recording_params(skipped=True)
        class Operation(object):

            @self.tape_recorder.operation()
            def execute(self):
                return 5

        instance = Operation()
        with patch.object(InMemoryTapeCassette, 'create_new_recording', wraps=self.tape_cassette.create_new_recording) \
                as intercepted:
            result = instance.execute()
            intercepted.assert_not_called()

        self.assertEqual(5, result)

    def test_input_interception_key_failure(self):

        class UnencodeableObject(object):
            def __getstate__(self):
                raise Exception()

        class Operation(object):

            @self.tape_recorder.operation()
            def execute(self):
                param = UnencodeableObject()
                value = self.input(param)
                return value

            @self.tape_recorder.intercept_input('input')
            def input(self, param):
                return 5

        instance = Operation()
        with patch.object(InMemoryTapeCassette, '_save_recording', wraps=self.tape_cassette._save_recording) \
                as intercepted_save, \
                patch.object(InMemoryTapeCassette, 'abort_recording', wraps=self.tape_cassette.abort_recording) \
                as intercepted_abort:
            result = instance.execute()
            intercepted_save.assert_not_called()
            intercepted_abort.assert_called()

        self.assertEqual(5, result)

    def test_input_interception_key_failure_during_playback(self):

        class UnencodeableObject(object):
            def __getstate__(self):
                raise Exception()

        class Operation(object):

            def __init__(self, raise_on_get=False):
                self.raise_on_get = raise_on_get

            @self.tape_recorder.operation()
            def execute(self):
                param = UnencodeableObject() if self.raise_on_get else object()
                value = self.input(param)
                return value

            @self.tape_recorder.intercept_input('input')
            def input(self, param):
                return 5

        instance = Operation()
        result = instance.execute()
        self.assertEqual(5, result)

        recording_id = self.tape_cassette.get_last_recording_id()
        with self.assertRaises(InputInterceptionKeyCreationError):
            self.tape_recorder.play(recording_id, playback_function=lambda recording: Operation(True).execute())

    def test_input_interception_key_missing_during_playback(self):

        class Operation(object):

            def __init__(self, input_key):
                self.input_key = input_key

            @self.tape_recorder.operation()
            def execute(self):
                value = self.input(self.input_key)
                return value

            @self.tape_recorder.intercept_input('input')
            def input(self, param):
                return 5

        instance = Operation('key1')
        result = instance.execute()
        self.assertEqual(5, result)

        recording_id = self.tape_cassette.get_last_recording_id()
        with self.assertRaises(RecordingKeyError):
            self.tape_recorder.play(recording_id, playback_function=lambda recording: Operation('key2').execute())

    def test_output_interception_key_missing_during_playback(self):

        class Operation(object):

            def __init__(self, output_method):
                self.output_method = output_method

            @self.tape_recorder.operation()
            def execute(self):
                value = getattr(self, self.output_method)()
                return value

            @self.tape_recorder.intercept_output('output_function')
            def output(self):
                return 5

            @self.tape_recorder.intercept_output('output_missing')
            def output_missing(self):
                return 5

        instance = Operation('output')
        result = instance.execute()
        self.assertEqual(5, result)

        recording_id = self.tape_cassette.get_last_recording_id()
        with self.assertRaises(RecordingKeyError):
            self.tape_recorder.play(recording_id,
                                    playback_function=lambda recording: Operation('output_missing').execute())

    def test_record_and_playback_basic_operation_metadata_extractor_raise_exception(self):
        def operation_metadata_extractor(*args, **kwargs):
            raise Exception('Meta exception')

        class Operation(object):

            @self.tape_recorder.operation(metadata_extractor=operation_metadata_extractor)
            def execute(self):
                return 5

        instance = Operation()
        result = instance.execute()
        self.assertEqual(5, result)

        recording_id = self.tape_cassette.get_last_recording_id()
        playback_result = self.tape_recorder.play(recording_id,
                                                  playback_function=lambda recording: Operation().execute())
        self._assert_playback_vs_recording(playback_result, result)

    def test_record_and_playback_basic_operation_save_recording_raise_exception(self):
        class Operation(object):

            @self.tape_recorder.operation()
            def execute(self):
                return 5

        with patch.object(InMemoryTapeCassette, '_save_recording', side_effect=Exception()):
            instance = Operation()
            result = instance.execute()
            self.assertEqual(5, result)

    def test_record_and_playback_basic_operation_no_parameters_raise_error(self):
        class Operation(object):

            @self.tape_recorder.operation()
            def execute(self):
                raise ValueError("Error")

        instance = Operation()
        with self.assertRaises(ValueError) as e:
            instance.execute()
        self.assertEqual("Error", str(e.exception))

        recording_id = self.tape_cassette.get_last_recording_id()
        playback_result = self.tape_recorder.play(recording_id,
                                                  playback_function=lambda recording: Operation().execute())
        operation_output = next(po for po in playback_result.playback_outputs
                                if TapeRecorder.OPERATION_OUTPUT_ALIAS in po.key)
        self.assertEqual(ValueError, type(operation_output.value['args'][0]))
        self.assertEqual("Error", str(operation_output.value['args'][0]))
        self.assertEqual(len(playback_result.recorded_outputs), len(playback_result.playback_outputs))
        self.assertGreater(playback_result.playback_duration, 0)
        self.assertGreater(playback_result.recorded_duration, 0)
        self.assertDictContainsSubset({TapeRecorder.EXCEPTION_IN_OPERATION: True},
                                      playback_result.original_recording.get_metadata())

    def test_record_and_playback_basic_operation_no_parameters_raise_unserializable_error(self):

        class UnserializableException(Exception):
            def __getstate__(self):
                raise Exception("Unserializable")

        class Operation(object):

            @self.tape_recorder.operation()
            def execute(self):
                raise UnserializableException("Error")

        instance = Operation()
        with self.assertRaises(UnserializableException) as e:
            instance.execute()
        self.assertEqual("Error", str(e.exception))

        recording_id = self.tape_cassette.get_last_recording_id()
        playback_result = self.tape_recorder.play(recording_id,
                                                  playback_function=lambda recording: Operation().execute())
        operation_output = next(po for po in playback_result.playback_outputs
                                if TapeRecorder.OPERATION_OUTPUT_ALIAS in po.key)
        self.assertEqual(UnserializableException, operation_output.value['args'][0]['error_type'])
        self.assertIn('Error', operation_output.value['args'][0]['error_repr'])
        self.assertEqual(len(playback_result.recorded_outputs), len(playback_result.playback_outputs))
        self.assertGreater(playback_result.playback_duration, 0)
        self.assertGreater(playback_result.recorded_duration, 0)
        self.assertDictContainsSubset({TapeRecorder.EXCEPTION_IN_OPERATION: True},
                                      playback_result.original_recording.get_metadata())

    def test_record_and_playback_basic_operation_data_interception_no_arguments_raise_exception(self):
        class Operation(object):

            def __init__(self, seed=0):
                self.seed = seed

            @self.tape_recorder.operation()
            def execute(self):
                try:
                    self.get_value()
                except Exception as ex:
                    return str(ex)

            @self.tape_recorder.intercept_input('input')
            def get_value(self):
                raise Exception(self.seed)

        instance = Operation(5)
        result = instance.execute()
        self.assertEqual('5', result)

        recording_id = self.tape_cassette.get_last_recording_id()
        playback_result = self.tape_recorder.play(recording_id,
                                                  playback_function=lambda recording: Operation().execute())
        self._assert_playback_vs_recording(playback_result, result)

    def test_record_and_playback_basic_operation_output_interception_no_arguments_raise_exception(self):
        class Operation(object):

            def __init__(self, seed=0):
                self.seed = seed

            @self.tape_recorder.operation()
            def execute(self):
                try:
                    self.output()
                except Exception as ex:
                    return str(ex)

            @self.tape_recorder.intercept_output('output_function')
            def output(self):
                raise Exception(self.seed)

        instance = Operation(5)
        result = instance.execute()
        self.assertEqual("5", result)

        recording_id = self.tape_cassette.get_last_recording_id()
        playback_result = self.tape_recorder.play(recording_id,
                                                  playback_function=lambda recording: Operation().execute())
        self._assert_playback_vs_recording(playback_result, result)

    def test_sampling_rate_class_level_decorator(self):

        @self.tape_recorder.recording_params(RecordingParameters(sampling_rate=0.1))
        class Operation(object):

            @self.tape_recorder.operation()
            def execute(self):
                return 5

        instance = Operation()
        with patch.object(InMemoryTapeCassette, 'save_recording', wraps=self.tape_cassette.create_new_recording) \
                as intercepted:
            calls = 100
            for __ in range(calls):
                result = instance.execute()
                self.assertEqual(5, result)

            call_ratio = float(intercepted.call_count)/calls
            print('Call ratio {}'.format(call_ratio))
            self.assertAlmostEqual(0.1, call_ratio, places=1)

    def test_operation_with_input_dict_as_key(self):
        class Operation(object):

            @self.tape_recorder.operation()
            def execute(self):
                # We are shuffling here hoping to get different internal order for the dict between calls but still
                # see we are consistent on input interception is these are equivalent dicts
                items = list(range(100))
                shuffle(items)
                argument = {'key{}'.format(i): i for i in items}
                return self.get_value(argument)

            @self.tape_recorder.intercept_input('input')
            def get_value(self, json_dict):
                return json_dict['key5']

        instance = Operation()
        result = instance.execute()
        self.assertEqual(5, result)

        recording_id = self.tape_cassette.get_last_recording_id()
        playback_result = self.tape_recorder.play(recording_id,
                                                  playback_function=lambda recording: Operation().execute())
        self._assert_playback_vs_recording(playback_result, result)

    def test_record_and_playback_operation_with_alias_param_resolver(self):
        class ValueCreator(object):
            def __init__(self, name, value):
                self.name = name
                self.value = value

            @self.tape_recorder.intercept_input('input.{name}', alias_params_resolver=lambda s: {'name': s.name})
            def get_value(self):
                return self.value

        class Operation(object):

            def __init__(self, seed=0):
                self.seed = seed

            @self.tape_recorder.operation()
            def execute(self):
                # We check that we are capturing different value interceptions even though we pass same arguments to
                # the intercepted method by using a param resolver that will provide unique interception key for each
                # invocation
                return self.get_value('a', self.seed) + self.get_value('b', self.seed * 2)

            @staticmethod
            def get_value(name, value):
                return ValueCreator(name, value).get_value()

        instance = Operation(5)
        result = instance.execute()
        self.assertEqual(15, result)

        recording_id = self.tape_cassette.get_last_recording_id()
        playback_result = self.tape_recorder.play(recording_id,
                                                  playback_function=lambda recording: Operation().execute())
        self._assert_playback_vs_recording(playback_result, result)

    def test_record_and_playback_operation_with_input_interception_data_handler(self):
        test_self = self

        class XorDataHandler(InputInterceptionDataHandler):

            def prepare_input_for_recording(self, interception_key, result, args, kwargs):
                test_self.assertIn('value_input', interception_key)
                test_self.assertEqual(5, result)
                test_self.assertEqual(('x', ), args[1:])
                test_self.assertEqual({'b': 'y'}, kwargs)
                return 5 ^ 3141

            def restore_input_from_recording(self, recorded_data, args, kwargs):
                # Applying same xor twice return original result
                return recorded_data ^ 3141

        class Operation(object):

            def __init__(self, seed=0):
                self.seed = seed

            @self.tape_recorder.operation()
            def execute(self):
                return self.get_value('x', b='y')

            @self.tape_recorder.intercept_input('value_input', data_handler=XorDataHandler())
            def get_value(self, a, b=None):
                return self.seed

        instance = Operation(5)
        result = instance.execute()
        self.assertEqual(5, result)

        recording_id = self.tape_cassette.get_last_recording_id()
        # Check the recording it self contains the modified value
        recording = self.tape_cassette.get_recording(recording_id)
        key = next(k for k in recording.get_all_keys() if 'value_input' in k)
        self.assertEqual(5 ^ 3141, recording.get_data(key)['value'])

        playback_result = self.tape_recorder.play(recording_id,
                                                  playback_function=lambda recording: Operation().execute())
        self._assert_playback_vs_recording(playback_result, result)

    def test_input_interception_data_handler_failure(self):

        class ErrorDataHandler(InputInterceptionDataHandler):

            def prepare_input_for_recording(self, interception_key, result, args, kwargs):
                raise Exception('error')

            def restore_input_from_recording(self, recorded_data, args, kwargs):
                pass

        class Operation(object):

            @self.tape_recorder.operation()
            def execute(self):
                return self.get_value() + self.get_value2()

            @self.tape_recorder.intercept_input('value_input', data_handler=ErrorDataHandler())
            def get_value(self):
                return 5

            @self.tape_recorder.intercept_input('value_input2')
            def get_value2(self):
                return 5

        instance = Operation()
        with patch.object(InMemoryTapeCassette, '_save_recording', wraps=self.tape_cassette._save_recording) \
                as intercepted:
            result = instance.execute()
            intercepted.assert_not_called()

        self.assertEqual(10, result)

    def test_record_and_playback_basic_operation_data_interception_with_all_arguments_ignored_as_key(self):
        class Operation(object):

            @self.tape_recorder.operation()
            def execute(self):
                arg1 = random()
                arg2 = random()
                val1 = self.get_value(arg1, b=arg2)
                return val1

            @self.tape_recorder.intercept_input('input', capture_args=[])
            def get_value(self, a, b=2):
                return 5

        instance = Operation()
        result = instance.execute()
        self.assertEqual(5, result)

        recording_id = self.tape_cassette.get_last_recording_id()
        playback_result = self.tape_recorder.play(recording_id,
                                                  playback_function=lambda recording: Operation().execute())
        self._assert_playback_vs_recording(playback_result, result)

    def test_record_and_playback_basic_operation_data_interception_with_false_capture_args(self):
        class Operation(object):

            @self.tape_recorder.operation()
            def execute(self):
                arg1 = random()
                arg2 = random()
                val1 = self.get_value(arg1, b=arg2)
                return val1

            @self.tape_recorder.intercept_input('input', capture_args=False)
            def get_value(self, a, b=2):
                return 5

        instance = Operation()
        result = instance.execute()
        self.assertEqual(5, result)

        recording_id = self.tape_cassette.get_last_recording_id()
        playback_result = self.tape_recorder.play(recording_id,
                                                  playback_function=lambda recording: Operation().execute())
        self._assert_playback_vs_recording(playback_result, result)

    def test_sampling_rate_with_enforced_recording(self):
        local_tape_recorder = self.tape_recorder
        test_self = self

        @self.tape_recorder.recording_params(RecordingParameters(sampling_rate=0.2))
        class Operation(object):

            @self.tape_recorder.operation()
            def execute(self):
                local_tape_recorder.force_sample_recording()
                test_self.assertTrue(local_tape_recorder.is_recording_sample_forced)
                return 5

        instance = Operation()
        with patch.object(InMemoryTapeCassette, '_save_recording', wraps=self.tape_cassette._save_recording) \
                as intercepted:
            calls = 10
            for __ in range(calls):
                result = instance.execute()
                self.assertEqual(5, result)

            call_ratio = float(intercepted.call_count)/calls
            print('Call ratio {}'.format(call_ratio))
            self.assertEqual(1, call_ratio)

    def test_sampling_rate_with_ignore_enforced_recording(self):
        local_tape_recorder = self.tape_recorder
        test_self = self

        @self.tape_recorder.recording_params(RecordingParameters(sampling_rate=0.1, ignore_enforced_sampling=True))
        class Operation(object):

            @self.tape_recorder.operation()
            def execute(self):
                local_tape_recorder.force_sample_recording()
                test_self.assertFalse(local_tape_recorder.is_recording_sample_forced)
                return 5

        instance = Operation()
        with patch.object(InMemoryTapeCassette, '_save_recording', wraps=self.tape_cassette._save_recording) \
                as intercepted_save,\
                patch.object(InMemoryTapeCassette, 'abort_recording', wraps=self.tape_cassette.abort_recording) \
                as intercepted_abort:
            calls = 100
            for __ in range(calls):
                result = instance.execute()
                self.assertEqual(5, result)

            expected_aborted = calls - intercepted_save.call_count
            call_ratio = float(intercepted_save.call_count) / calls
            print('Call ratio {}'.format(call_ratio))
            self.assertAlmostEqual(0.1, call_ratio, places=1)
            self.assertEqual(expected_aborted, intercepted_abort.call_count)

    def test_record_and_playback_basic_operation_data_interception_with_specific_arguments_ignored_as_key(self):
        class Operation(object):

            @self.tape_recorder.operation()
            def execute(self):
                # This is random to see that capture args is only capturing the requested arg index
                arg1 = random()
                arg2 = 'b'
                arg3 = 'c'
                # This is random to see that capture args is only capturing the requested arg kwargs
                arg4 = random()
                arg5 = 'e'
                val1 = self.get_value(arg1, arg2, c=arg3, d=arg4, e=arg5)
                return val1

            @self.tape_recorder.intercept_input('input', capture_args=[CapturedArg(2, 'b'), CapturedArg(3, 'c'),
                                                                       CapturedArg(None, 'e'), CapturedArg(None, 'f')])
            def get_value(self, a, b, c, d=None, e=None, f=None):
                return 5

        instance = Operation()
        result = instance.execute()
        self.assertEqual(5, result)

        recording_id = self.tape_cassette.get_last_recording_id()
        playback_result = self.tape_recorder.play(recording_id,
                                                  playback_function=lambda recording: Operation().execute())
        self._assert_playback_vs_recording(playback_result, result)

    def test_record_and_playback_operation_with_output_interception_data_handler(self):
        test_self = self

        class XorDataHandler(OutputInterceptionDataHandler):

            def prepare_output_for_recording(self, interception_key, args, kwargs):
                test_self.assertIn('output_function', interception_key)
                test_self.assertEqual((5, ), args)
                test_self.assertEqual({'arg': 'a'}, kwargs)
                return args[0] ^ 3141

            def restore_output_from_recording(self, recorded_data):
                # Applying same xor twice return original result
                return recorded_data ^ 3141

        class Operation(object):

            @self.tape_recorder.operation()
            def execute(self):
                return self.output(5, arg='a')

            @self.tape_recorder.intercept_output('output_function', data_handler=XorDataHandler())
            def output(self, value, arg=None):
                return value

        instance = Operation()
        result = instance.execute()
        self.assertEqual(5, result)

        recording_id = self.tape_cassette.get_last_recording_id()
        # Check the recording it self contains the modified value
        recording = self.tape_cassette.get_recording(recording_id)
        key = next(k for k in recording.get_all_keys() if 'output_function' in k)
        self.assertEqual(5 ^ 3141, recording.get_data(key))
        playback_result = self.tape_recorder.play(recording_id,
                                                  playback_function=lambda recording: Operation().execute())

        self._assert_playback_vs_recording(playback_result, result)

        self.assertEqual(result, XorDataHandler().restore_output_from_recording(
            playback_result.playback_outputs[0].value))

    def test_output_interception_data_handler_failure(self):

        class ErrorDataHandler(OutputInterceptionDataHandler):

            def prepare_output_for_recording(self, interception_key, args, kwargs):
                raise Exception('error')

            def restore_output_from_recording(self, recorded_data):
                pass

        class Operation(object):

            @self.tape_recorder.operation()
            def execute(self):
                return self.output(5) + self.output2(5)

            @self.tape_recorder.intercept_output('output_function', data_handler=ErrorDataHandler())
            def output(self, value):
                return value

            @self.tape_recorder.intercept_output('output_function2')
            def output2(self, value):
                return value

        instance = Operation()
        with patch.object(InMemoryTapeCassette, '_save_recording', wraps=self.tape_cassette._save_recording) \
                as intercepted:
            result = instance.execute()
            intercepted.assert_not_called()

        self.assertEqual(10, result)

    def test_record_and_playback_basic_operation_data_interception_inside_interception(self):
        class Operation(object):

            def __init__(self, seed=0):
                self.seed = seed

            @self.tape_recorder.operation()
            def execute(self):
                return self.get_value()

            @self.tape_recorder.intercept_input('input')
            def get_value(self):
                return self.get_inner_value()

            @self.tape_recorder.intercept_input('inner_input')
            def get_inner_value(self):
                return self.seed

        instance = Operation(5)
        result = instance.execute()
        self.assertEqual(5, result)

        recording_id = self.tape_cassette.get_last_recording_id()
        playback_result = self.tape_recorder.play(recording_id,
                                                  playback_function=lambda recording: Operation().execute())
        self._assert_playback_vs_recording(playback_result, result)

        # Check the recording it self does not contain the inner interception
        recording = self.tape_cassette.get_recording(recording_id)
        self.assertIsNone(next((k for k in recording.get_all_keys() if 'inner_input' in k), None))

    def test_record_and_playback_basic_operation_new_output_added_post_recording(self):
        class OperationOld(object):

            @self.tape_recorder.operation()
            def execute(self):
                return self.output(5)

            @self.tape_recorder.intercept_output('output_function')
            def output(self, value):
                return value

        class OperationNew(object):

            @self.tape_recorder.operation()
            def execute(self):
                value = self.output(5)
                self.output_new(value)
                return value

            @self.tape_recorder.intercept_output('output_function')
            def output(self, value):
                return value

            @self.tape_recorder.intercept_output('output_new_function', fail_on_no_recorded_result=False)
            def output_new(self, value):
                return value

        instance = OperationOld()
        result = instance.execute()
        self.assertEqual(5, result)

        recording_id = self.tape_cassette.get_last_recording_id()
        playback_result = self.tape_recorder.play(recording_id,
                                                  playback_function=lambda recording: OperationNew().execute())

        self.assertEqual({'args': [5], 'kwargs': {}}, playback_result.playback_outputs[0].value)
        self.assertEqual({'args': [5], 'kwargs': {}}, playback_result.playback_outputs[1].value)
        self.assertNotEqual(playback_result.playback_outputs[0].key, playback_result.playback_outputs[1].key)
        self.assertIn('output_function', playback_result.playback_outputs[0].key)
        self.assertIn('output_new_function', playback_result.playback_outputs[1].key)

    def test_record_and_playback_basic_operation_new_output_added_post_recording_with_default_return_values(self):
        class OperationOld(object):

            @self.tape_recorder.operation()
            def execute(self):
                return self.output(5)

            @self.tape_recorder.intercept_output('output_function')
            def output(self, value):
                return value

        class OperationNew(object):

            @self.tape_recorder.operation()
            def execute(self):
                value = self.output(5)
                val1, val2 = self.output_new(value)
                return val1 + val2

            @self.tape_recorder.intercept_output('output_function')
            def output(self, value):
                return value

            @self.tape_recorder.intercept_output('output_new_function', fail_on_no_recorded_result=False,
                                                 default_result_when_not_recorded=(2, 4))
            def output_new(self, value):
                return value, value * 2

        instance = OperationOld()
        result = instance.execute()
        self.assertEqual(5, result)

        recording_id = self.tape_cassette.get_last_recording_id()
        playback_result = self.tape_recorder.play(recording_id,
                                                  playback_function=lambda recording: OperationNew().execute())

        self.assertEqual({'args': [5], 'kwargs': {}}, playback_result.playback_outputs[0].value)
        self.assertEqual({'args': [5], 'kwargs': {}}, playback_result.playback_outputs[1].value)
        self.assertNotEqual(playback_result.playback_outputs[0].key, playback_result.playback_outputs[1].key)
        self.assertIn('output_function', playback_result.playback_outputs[0].key)
        self.assertIn('output_new_function', playback_result.playback_outputs[1].key)
        operation_output = next(po for po in playback_result.playback_outputs
                                if TapeRecorder.OPERATION_OUTPUT_ALIAS in po.key)
        self.assertEqual(6, operation_output.value['args'][0])

    def test_current_recording_id(self):
        tape_recorder = self.tape_recorder

        class Operation(object):

            @self.tape_recorder.operation()
            def execute(self):
                return tape_recorder.current_recording_id

        instance = Operation()
        result = instance.execute()
        recording_id = self.tape_cassette.get_last_recording_id()
        self.assertEqual(recording_id, result)

        playback_result = self.tape_recorder.play(recording_id,
                                                  playback_function=lambda recording: Operation().execute())

        operation_output = next(po for po in playback_result.playback_outputs
                                if TapeRecorder.OPERATION_OUTPUT_ALIAS in po.key)
        self.assertEqual(recording_id, operation_output.value['args'][0])

        self.tape_recorder.disable_recording()
        instance = Operation()
        self.assertIsNone(instance.execute())

    def test_record_and_playback_record_and_play_data(self):
        tape_recorder = self.tape_recorder

        class Operation(object):

            @self.tape_recorder.operation()
            def execute(self):
                data = tape_recorder.play_data('data')
                tape_recorder.record_data('data', data)
                if tape_recorder.in_playback_mode:
                    return data

                data = 5
                tape_recorder.record_data('data', data)
                return data

        instance = Operation()
        result = instance.execute()
        self.assertEqual(5, result)

        recording_id = self.tape_cassette.get_last_recording_id()
        with patch.object(TapeRecorder, 'play_data', wraps=tape_recorder.play_data) as wrapped:
            playback_result = self.tape_recorder.play(recording_id,
                                                      playback_function=lambda recording: Operation().execute())
        self._assert_playback_vs_recording(playback_result, result)
        wrapped.assert_called()

    @patch('playback.tape_recorder.pickle_copy', side_effect=pickle_copy)
    def test_record_and_playback_basic_operation_data_interception_copy_data(self, wrapped_encode):
        @self.tape_recorder.recording_params(RecordingParameters(copy_data_on_intercepion=True))
        class Operation(object):

            def __init__(self, seed=0):
                self.seed = seed

            @self.tape_recorder.operation()
            def execute(self):
                return self.get_value()

            @self.tape_recorder.intercept_input('input')
            def get_value(self):
                return self.seed

        instance = Operation(5)
        result = instance.execute()
        self.assertEqual(5, result)

        recording_id = self.tape_cassette.get_last_recording_id()
        playback_result = self.tape_recorder.play(recording_id,
                                                  playback_function=lambda recording: Operation().execute())
        wrapped_encode.assert_called_with(5)
        self._assert_playback_vs_recording(playback_result, result)

    @patch('playback.tape_recorder.pickle_copy')
    def test_record_and_playback_basic_operation_data_interception_copy_data_exception(self, wrapped_decode):

        @self.tape_recorder.recording_params(RecordingParameters(copy_data_on_intercepion=True))
        class Operation(object):

            def __init__(self, seed=0):
                self.seed = seed

            @self.tape_recorder.operation()
            def execute(self):
                return self.get_value()

            @self.tape_recorder.intercept_input('input')
            def get_value(self):
                return self.seed

        def cannot(input):
            raise Exception('cannot')

        wrapped_decode.side_effect = cannot

        instance = Operation(5)
        result = instance.execute()
        self.assertEqual(5, result)

        recording_id = self.tape_cassette.get_last_recording_id()
        playback_result = self.tape_recorder.play(recording_id,
                                                  playback_function=lambda recording: Operation().execute())
        wrapped_decode.assert_called()
        self._assert_playback_vs_recording(playback_result, result)

    def test_intercept_input_activate_original_method_on_missing(self):
        class OperationOld(object):

            def __init__(self, seed=0):
                self.seed = seed

            @self.tape_recorder.operation()
            def execute(self):
                return self.get_value()

            def get_value(self):
                return self.seed

        class OperationNew(object):

            def __init__(self, seed=0):
                self.seed = seed

            @self.tape_recorder.operation()
            def execute(self):
                return self.get_value()

            @self.tape_recorder.intercept_input('input', run_intercepted_when_missing=True)
            def get_value(self):
                return self.seed

        instance = OperationOld(5)
        result = instance.execute()
        self.assertEqual(5, result)

        recording_id = self.tape_cassette.get_last_recording_id()
        playback_result = self.tape_recorder.play(recording_id,
                                                  playback_function=lambda recording: OperationNew(5).execute())
        self._assert_playback_vs_recording(playback_result, result)



from __future__ import absolute_import
import logging
import os
import base64
import six
from playback.utils.timing_utils import Timed

_logger = logging.getLogger(__name__)


class FileInterception(object):

    ABOVE_LIMIT_CONTENT = six.b('above interception limit')

    def __init__(self, file_path_arg_index, file_path_arg_name, intercepted_size_limit=None):
        """
        :param file_path_arg_index: The index inside arguments that points to the file path that needs to be intercepted
        :type file_path_arg_index: int
        :param file_path_arg_name: The name of the argument that points to the file path that needs to be intercepted
        :type file_path_arg_name: str
        :param intercepted_size_limit: Files that their size in MB is above this limit will not be intercepted
        :type intercepted_size_limit: float
        """
        self.file_path_arg_index = file_path_arg_index
        self.file_path_arg_name = file_path_arg_name
        self.intercepted_size_limit = self._calculate_max_intercepted_size_limit(intercepted_size_limit)

    @staticmethod
    def _mb_size(size):
        """
        :return: Size in MB
        :rtype: float
        """
        return size / (1024.0 * 1024.0)

    @staticmethod
    def _calculate_max_intercepted_size_limit(intercepted_size_limit):
        """
        :param intercepted_size_limit: Files that their size in MB is above this limit will not be intercepted
        :type intercepted_size_limit: float
        :return: Max intercepted file size
        :rtype: float
        """
        if intercepted_size_limit is not None:
            return intercepted_size_limit

        return int(float(os.getenv('PLAYBACK_INTERCEPTED_FILE_SIZE_LIMIT', "500")))

    def _get_file_path(self, args, kwargs):
        """
        :param args: Invocation args
        :type args: tuple
        :param kwargs: Invocation kwargs
        :type kwargs: dict
        :return: Intercepted file path from invocation parameters
        :rtype: basestring
        """
        file_path = kwargs.get(self.file_path_arg_name)
        if not file_path:
            file_path = args[self.file_path_arg_index]
        return file_path

    def _intercept_file(self, args, kwargs):
        """
        Record the file passed in the args/kwargs
        :param args: Invocation args
        :type args: tuple
        :param kwargs: Invocation kwargs
        :type kwargs: dict
        :return: Intercepted result in a form that should be saved in the recording
        :rtype: dict[str, str]
        """
        file_path = self._get_file_path(args, kwargs)

        if self._is_file_above_size_limit(file_path):
            return self._above_limit_result(file_path)

        _logger.info(u'Reading intercepted file ({})'.format(file_path))

        with Timed() as timed:
            with open(file_path, "rb") as f:
                content = f.read()
            _logger.info(u'Done reading content size is {:.2f}MB ({})'.format(self._mb_size(len(content)), file_path))
            result = self._serialize_file(content, file_path)

        _logger.info(u'Done preparing file for recording with in {:.2f}s ({})'.format(timed.duration, file_path))
        return result

    def _is_file_above_size_limit(self, file_path):
        """
        Checks if intercepted file is above the configured size limit for interception
        :param file_path: Path
        :type file_path: basestring
        :return: Whether the file size is above interception limit
        :rtype: bool
        """
        if self.intercepted_size_limit is not None:
            file_size_in_mb = self._mb_size(os.path.getsize(file_path))
            if file_size_in_mb > self.intercepted_size_limit:
                _logger.info(u'Intercepted file is {:.2f}MB which is above interception limit of {:.2f}MB, '
                             u'ignoring content in file {}'.format(
                                 file_size_in_mb, self.intercepted_size_limit, file_path))
                return True
        return False

    @staticmethod
    def _above_limit_result(file_path):
        """
        Returns file above limit result (removing the content)
        :param file_path: Path of file
        :type file_path: basestring
        :return: File above limit result (removing the content)
        :rtype: dict[str, str]
        """
        return {
            'file_path': file_path,
            'file_content': FileInterception.ABOVE_LIMIT_CONTENT
        }

    @staticmethod
    def _serialize_file(content, file_path):
        """
        Serialize the file content prepared for recording
        :param content: File content
        :type content: basestring
        :param file_path: Path of file
        :type file_path: basestring
        :return: Serialized form of file
        :rtype: dict[str, str]
        """

        if six.PY2:
            encoded_content = content.encode('base64')
        else:
            encoded_content = base64.b64encode(content)

        return {
            'file_path': file_path,
            'file_content': encoded_content
        }

    @staticmethod
    def _deserialize_file(serialized_file):
        """
        Deserialize the content into file path and file content
        :param serialized_file: Serialized form of file
        :type serialized_file: dict[str, str]
        :return: File path and file content
        :rtype: str, str
        """
        file_content = serialized_file['file_content']

        if file_content != FileInterception.ABOVE_LIMIT_CONTENT:
            if six.PY2:
                file_content = file_content.decode('base64')
            else:
                file_content = base64.b64decode(file_content)

        return serialized_file['file_path'], file_content



import six
from flask import request
from flask_restplus import Resource

from examples.flask.playback_context import tape_recorder


class ContentBasedService(Resource):
    @tape_recorder.intercept_input('content_based_service.request_data')
    def _get_request_data(self):
        return request.json

    @tape_recorder.intercept_input('content_based_service.get_url_content')
    def _get_url_content(self, url):
        print("Fetching content from url: {}".format(url))
        if six.PY3:
            import urllib.request
            return urllib.request.urlopen(url).read()
        else:
            import urllib2
            return urllib2.urlopen(url).read().decode("utf8")

    @tape_recorder.intercept_output('content_based_service.persist_invocation')
    def _persist_invocation(self, url, result):
        # Mimic a persist operation for extra optional output, in this case we are not really doing anything with it but
        # it could be a data base call
        print("Persisting operation result - url: {}, result: {}".format(url, result))


class ContentLengthEndpoint(ContentBasedService):

    @tape_recorder.operation()
    def post(self):
        # First input - read the data from the request
        request_data = self._get_request_data()
        url = request_data['url']
        # Second input - read the data from the request url. The url passed to the input is part of the intercepted key,
        # if the code would change and upon playback the url would be different due to code changes, the call to
        # _get_url_content will fail with missing key
        content = self._get_url_content(url)
        content_length = len(content)
        self._persist_invocation(url, content_length)
        return content_length


class ContentFirstCharsEndpoint(ContentBasedService):

    @tape_recorder.operation()
    def post(self):
        # First input - read the data from the request
        request_data = self._get_request_data()
        url = request_data['url']
        length = request_data['length']
        # Second input - read the data from the request url. The url passed to the input is part of the intercepted key,
        # if the code would change and upon playback the url would be different due to code changes, the call to
        # _get_url_content will fail with missing key
        content = self._get_url_content(url)
        content_chars = content[:length]
        self._persist_invocation(url, content_chars)
        return content_chars



from abc import ABCMeta, abstractmethod


class TapeCassette(object):
    """
    An abstract class that acts as a storage driver for TapeRecorder to store and fetch recordings
    """

    __metaclass__ = ABCMeta

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @abstractmethod
    def get_recording(self, recording_id):
        """
        Get recording stored with the given id
        :param recording_id: If of recording to fetch
        :type recording_id: basestring
        :return: Recording of the given id
        :rtype: playback.recording.Recording
        :raises: playback.exceptions.NoSuchRecording
        """
        pass

    def get_recording_metadata(self, recording_id):
        """
        Get recording's metadata stored with the given id
        :param recording_id: If of recording to fetch
        :type recording_id: basestring
        :return: Recording of the given id
        :rtype: dict
        :raises: playback.exceptions.NoSuchRecording
        """
        return self.get_recording(recording_id).get_metadata()

    @abstractmethod
    def create_new_recording(self, category):
        """
        Creates a new recording object
        :param category: A category to classify the recording in (e.g operation class) (serializable)
        :type category: Any
        :return: Creates a new recording object
        :rtype: playback.recording.Recording
        """
        pass

    def abort_recording(self, recording=None):
        # pylint: disable=no-self-use
        """
        Aborts given recording without saving it
        :param recording: Recording to abort
        :type recording: playback.recording.Recording
        """
        recording.close()

    def save_recording(self, recording):
        """
        Saves given recording
        :param recording: Recording to save
        :type recording: playback.recording.Recording
        """
        self._save_recording(recording)
        recording.close()

    @abstractmethod
    def _save_recording(self, recording):
        """
        Saves given recording
        :param recording: Recording to save
        :type recording: playback.recording.Recording
        """
        pass

    @abstractmethod
    def iter_recording_ids(self, category, start_date=None, end_date=None, metadata=None, limit=None):
        """
        Creates an iterator of recording ids matching the given search parameters
        :param category: Recordings category
        :type category: str
        :param start_date: Optional recording start date (need to be given in utc time)
        :type start_date: datetime.datetime
        :param end_date: Optional recording end date (need to be given in utc time)
        :type end_date: datetime.datetime
        :param metadata: Optional metadata values to filter by
        :type metadata: dict
        :param limit: Optional limit on number of ids to fetch
        :type limit: int
        :return: Iterator of recording ids matching the given parameters
        :rtype: collections.Iterator[str]
        """
        pass

    def iter_recordings_metadata(self, category, start_date=None, end_date=None, metadata=None, limit=None):
        """
        Creates an iterator of recordings metadata matching the given search parameters
        :param category: Recordings category
        :type category: str
        :param start_date: Optional recording start date (need to be given in utc time)
        :type start_date: datetime.datetime
        :param end_date: Optional recording end date (need to be given in utc time)
        :type end_date: datetime.datetime
        :param metadata: Optional metadata values to filter by
        :type metadata: dict
        :param limit: Optional limit on number of ids to fetch
        :type limit: int
        :return: Iterator of recording ids matching the given parameters
        :rtype: collections.Iterator[dict]
        """
        for recording_id in self.iter_recording_ids(
                category, start_date, end_date, metadata, limit):
            yield self.get_recording_metadata(recording_id)

    @abstractmethod
    def extract_recording_category(self, recording_id):
        """
        :param recording_id: Recording id to extract category from
        :type recording_id: str
        :return: Recording's category
        :rtype: str
        """
        pass

    def close(self):
        """
        Close this cassette and release any underlying resources
        """
        pass



from playback.recordings.memory.memory_recording import MemoryRecording
from time import sleep

from playback.tape_cassettes.in_memory.in_memory_tape_cassette import InMemoryTapeCassette


class DelayedMemoryRecording(MemoryRecording):

    def __init__(self, delay):
        super(DelayedMemoryRecording, self).__init__()
        self.delay = delay

    def _set_data(self, key, value):
        """
        Sets data in the recording
        :param key: data key
        :type key: basestring
        :param value: data value (serializable)
        :type value: Any
        """
        sleep(self.delay)
        return super(DelayedMemoryRecording, self)._set_data(key, value)

    def _add_metadata(self, metadata):
        """
        :param metadata: Metadata to add to the recording
        :type metadata: dict
        """
        sleep(self.delay)
        return super(DelayedMemoryRecording, self)._add_metadata(metadata)


class DelayedInMemoryTapeCassette(InMemoryTapeCassette):
    """
    Add artificial delay of recording to in memory tape cassette, used for testing asynchronous implementations
    """

    def __init__(self, delay, *args, **kwargs):
        """
        :param delay: How much time to artificially delay recording invocations
        :type delay: float
        """
        super(DelayedInMemoryTapeCassette, self).__init__(*args, **kwargs)
        self.delay = delay

    def create_new_recording(self, category):
        """
        :param category: A category to classify the recording in (e.g operation class) (serializable)
        :type category: Any
        :return: Creates a new recording object
        :rtype: playback.recording.Recording
        """
        return DelayedMemoryRecording(self.delay)

    def _save_recording(self, recording):
        """
        Saves given recording
        :param recording: Recording to save
        :type recording: playback.recording.Recording
        """
        sleep(self.delay)
        super(DelayedInMemoryTapeCassette, self)._save_recording(recording)



from random import shuffle


class RecordingLookupProperties(object):
    def __init__(self, start_date, end_date=None, metadata=None, limit=None, random_sample=False):
        """
        :param start_date: Earliest date of recording
        :type start_date: datetime.datetime
        :param end_date: Latest date of recording
        :type end_date: datetime.datetime
        :param metadata: Optional metadata to filter recording by
        :type metadata: dict
        :param limit: Limit the number of recordings to fetch
        :type limit: int
        :param random_sample: True/False collect using random.shuffle (use random.seed to change selection)
        :type random_sample: boolean
        """
        self.start_date = start_date
        self.end_date = end_date
        self.metadata = metadata
        self.limit = limit
        self.random_sample = random_sample


def find_matching_recording_ids(tape_recorder, category, lookup_properties):
    """
    :param tape_recorder: Tape cassette holding the recordings
    :type tape_recorder: playback.tape_recorder.TapeRecorder
    :param category: Recording category
    :type category: basestring
    :param lookup_properties: Recording lookup properties
    :type lookup_properties: RecordingLookupProperties
    :return: Iterator of recording ids based on lookup parameters
    :rtype: collections.Iterator[str]
    """
    recording_ids = tape_recorder.tape_cassette.iter_recording_ids(
            category, start_date=lookup_properties.start_date, end_date=lookup_properties.end_date,
            metadata=lookup_properties.metadata,
            limit=(None if lookup_properties.random_sample else lookup_properties.limit))

    if lookup_properties.random_sample:
        recording_ids = list(recording_ids)
        shuffle(recording_ids)
        recording_ids = iter(recording_ids[:lookup_properties.limit])

    return recording_ids



from abc import ABCMeta, abstractmethod


class EqualizerTuning(object):
    def __init__(self, playback_function, result_extractor, comparator, comparison_data_extractor=None):
        """
        :param playback_function: A function that plays back the operation using the recording in the given id
        :type playback_function: function
        :param result_extractor: Extracts result from the recording and playback
        :type result_extractor: function
        :param comparison_data_extractor: Extracts optional data from the recording that will be passed to the
        comparator
        :type comparison_data_extractor: function
        :param comparator: A function use to create the equality status by comparing the expected vs actual result
        :type comparator: function
        """
        self.playback_function = playback_function
        self.result_extractor = result_extractor
        self.comparator = comparator
        self.comparison_data_extractor = comparison_data_extractor


class EqualizerTuner(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def create_category_tuning(self, category):
        """
        :param category: Category
        :type category: basestring
        :return: Tuning for category
        :rtype: EqualizerTuning
        """
        pass



# p3ready
from __future__ import absolute_import
import unittest
from time import sleep, time

from contextlib2 import suppress
from datetime import datetime, timedelta
from mock import patch
import random

from parameterized import parameterized

from playback.studio.recordings_lookup import find_matching_recording_ids, RecordingLookupProperties

from playback.studio.equalizer import Equalizer, EqualityStatus, ComparatorResult, CompareExecutionConfig
from playback.tape_recorder import TapeRecorder
from playback.tape_cassettes.in_memory.in_memory_tape_cassette import InMemoryTapeCassette
from six.moves import range


def exact_comparator(recorded_result, playback_result):
    if type(recorded_result) != type(playback_result):
        return EqualityStatus.Different
    diff = abs(recorded_result - playback_result)
    if diff == 0:
        return EqualityStatus.Equal
    if diff <= 10:
        return EqualityStatus.Different
    return EqualityStatus.Failed


def exact_comparator_with_message(recorded_result, playback_result):
    result = exact_comparator(recorded_result, playback_result)
    return ComparatorResult(result, result.name)


def return_value_result_extractor(outputs):
    return next(o.value['args'][0] for o in outputs if TapeRecorder.OPERATION_OUTPUT_ALIAS in o.key)


class TestEqualizer(unittest.TestCase):

    def setUp(self):
        self.tape_cassette = InMemoryTapeCassette()
        self.tape_recorder = TapeRecorder(self.tape_cassette)
        self.tape_recorder.enable_recording()

    def tearDown(self):
        self.tape_cassette.close()

    @parameterized.expand([("compare_in_dedicated_process", True),
                           ("playback_same_process", False),
                           ])
    def test_equal_comparison(self, name, compare_in_dedicated_process):

        class Operation(object):
            def __init__(self, value=None, multiply_input=1):
                self._value = value
                self.multiply_input = multiply_input

            @property
            @self.tape_recorder.intercept_input('input')
            def input(self):
                return self._value

            @self.tape_recorder.operation()
            def execute(self):
                return self.input * self.multiply_input

        Operation(3).execute()
        Operation(4).execute()
        Operation(5).execute()

        playback_counter = [0]

        def playback_function(recording):
            playback_counter[0] += 1
            if playback_counter[0] == 2:
                operation = Operation(multiply_input=2)
            elif playback_counter[0] == 3:
                operation = Operation(multiply_input=100)
            else:
                operation = Operation()
            return operation.execute()

        def player(recording_id):
            return self.tape_recorder.play(recording_id, playback_function)

        with patch.object(InMemoryTapeCassette, 'iter_recording_ids',
                          wraps=self.tape_cassette.iter_recording_ids) as mock:
            start_date = datetime.utcnow() - timedelta(hours=1)
            end_date = datetime.utcnow() + timedelta(hours=1)
            playable_recordings = find_matching_recording_ids(
                self.tape_recorder,
                category=Operation.__name__,
                lookup_properties=RecordingLookupProperties(start_date=start_date, end_date=end_date),
            )

            runner = Equalizer(playable_recordings, player, result_extractor=return_value_result_extractor,
                               comparator=exact_comparator,
                               compare_execution_config=CompareExecutionConfig(
                                   keep_results_in_comparison=True,
                                   compare_in_dedicated_process=compare_in_dedicated_process
                               ))

            with patch.object(Equalizer, '_play_and_compare_recording',
                              wraps=runner._play_and_compare_recording) as wrapped:
                comparison = list(runner.run_comparison())
            if compare_in_dedicated_process:
                wrapped.assert_not_called()
            else:
                wrapped.assert_called()

        self.assertEqual(EqualityStatus.Equal, comparison[0].comparator_status.equality_status)
        self.assertEqual(EqualityStatus.Different, comparison[1].comparator_status.equality_status)
        self.assertEqual(EqualityStatus.Failed, comparison[2].comparator_status.equality_status)

        self.assertIsNotNone(comparison[0].recording_id)
        self.assertEqual(comparison[0].playback.original_recording.id, comparison[0].recording_id)
        self.assertIsNotNone(comparison[1].recording_id)
        self.assertEqual(comparison[1].playback.original_recording.id, comparison[1].recording_id)
        self.assertIsNotNone(comparison[2].recording_id)
        self.assertEqual(comparison[2].playback.original_recording.id, comparison[2].recording_id)

        self.assertEqual(3, comparison[0].expected)
        self.assertEqual(4, comparison[1].expected)
        self.assertEqual(5, comparison[2].expected)

        self.assertEqual(3, comparison[0].actual)
        self.assertEqual(8, comparison[1].actual)
        self.assertEqual(500, comparison[2].actual)

        mock.assert_called_with(Operation.__name__, start_date=start_date, end_date=end_date, limit=None, metadata=None)
        self.assertGreaterEqual(comparison[0].playback.playback_duration, 0)
        self.assertGreaterEqual(comparison[0].playback.recorded_duration, 0)

    @parameterized.expand([("compare_in_dedicated_process", True),
                           ("playback_same_process", False),
                           ])
    def test_with_exception_comparison(self, name, compare_in_dedicated_process):

        class Operation(object):
            def __init__(self, value=None, raise_error=None):
                self._value = value
                self.raise_error = raise_error

            @property
            @self.tape_recorder.intercept_input('input')
            def input(self):
                return self._value

            @self.tape_recorder.operation()
            def execute(self, value=None):
                if self.raise_error:
                    raise Exception("error")
                if value is not None:
                    return value
                return self.input

        Operation(3).execute()
        Operation(4).execute()
        with suppress(Exception):
            Operation(raise_error=True).execute()

        playback_counter = [0]

        def playback_function(recording):
            playback_counter[0] += 1
            if playback_counter[0] == 2:
                operation = Operation(raise_error=True)
            elif playback_counter[0] == 3:
                return Operation().execute(5)
            else:
                operation = Operation()
            return operation.execute()

        def player(recording_id):
            return self.tape_recorder.play(recording_id, playback_function)

        start_date = datetime.utcnow() - timedelta(hours=1)
        playable_recordings = find_matching_recording_ids(
            self.tape_recorder,
            category=Operation.__name__,
            lookup_properties=RecordingLookupProperties(start_date=start_date),
        )
        runner = Equalizer(
            playable_recordings, player, result_extractor=return_value_result_extractor,
            comparator=exact_comparator, compare_execution_config=CompareExecutionConfig(
                keep_results_in_comparison=True,
                compare_in_dedicated_process=compare_in_dedicated_process
            )
        )

        comparison = list(runner.run_comparison())

        self.assertEqual(EqualityStatus.Equal, comparison[0].comparator_status.equality_status)
        self.assertEqual(EqualityStatus.Different, comparison[1].comparator_status.equality_status)
        self.assertEqual(EqualityStatus.Different, comparison[2].comparator_status.equality_status)

        self.assertEqual(3, comparison[0].expected)
        self.assertEqual(4, comparison[1].expected)
        self.assertIsInstance(comparison[2].expected, Exception)
        self.assertTrue(comparison[1].actual_is_exception)
        self.assertFalse(comparison[1].expected_is_exception)

        self.assertEqual(3, comparison[0].actual)
        self.assertIsInstance(comparison[1].actual, Exception)
        self.assertEqual(5, comparison[2].actual)
        self.assertFalse(comparison[2].actual_is_exception)
        self.assertTrue(comparison[2].expected_is_exception)

    @parameterized.expand([("compare_in_dedicated_process", True),
                           ("playback_same_process", False),
                           ])
    def test_with_failure_post_playback_comparison(self, name, compare_in_dedicated_process):

        class Operation(object):
            def __init__(self, value=None, multiply_input=1):
                self._value = value
                self.multiply_input = multiply_input

            @property
            @self.tape_recorder.intercept_input('input')
            def input(self):
                return self._value

            @self.tape_recorder.operation()
            def execute(self):
                return self.input * self.multiply_input

        Operation(3).execute()
        Operation(4).execute()
        Operation(5).execute()

        playback_counter = [0]

        def playback_function(recording):
            playback_counter[0] += 1
            if playback_counter[0] == 2:
                operation = Operation(multiply_input=2)
            elif playback_counter[0] == 3:
                operation = Operation(multiply_input=100)
            else:
                operation = Operation()
            return operation.execute()

        def result_extractor(outputs):
            value = next(o.value['args'][0] for o in outputs if TapeRecorder.OPERATION_OUTPUT_ALIAS in o.key)
            if value == 4:
                raise Exception('extract error')
            return value

        def player(recording_id):
            return self.tape_recorder.play(recording_id, playback_function)

        start_date = datetime.utcnow() - timedelta(hours=1)
        end_date = datetime.utcnow() + timedelta(hours=1)
        playable_recordings = find_matching_recording_ids(
            self.tape_recorder,
            category=Operation.__name__,
            lookup_properties=RecordingLookupProperties(start_date=start_date, end_date=end_date),
        )

        runner = Equalizer(playable_recordings, player, result_extractor=result_extractor,
                           comparator=exact_comparator,
                           compare_execution_config=CompareExecutionConfig(
                               keep_results_in_comparison=False,
                               compare_in_dedicated_process=compare_in_dedicated_process
                           ))

        comparison = list(runner.run_comparison())

        self.assertEqual(EqualityStatus.Equal, comparison[0].comparator_status.equality_status)
        self.assertEqual(EqualityStatus.EqualizerFailure, comparison[1].comparator_status.equality_status)
        self.assertEqual('extract error', comparison[1].comparator_status.message)
        self.assertIsNotNone(comparison[1].recording_id)
        self.assertIsNotNone(comparison[1].playback)
        self.assertEqual(EqualityStatus.Failed, comparison[2].comparator_status.equality_status)

        self.assertIsNone(comparison[0].expected)
        self.assertIsNone(comparison[0].actual)

    @parameterized.expand([("compare_in_dedicated_process", True),
                           ("playback_same_process", False),
                           ])
    def test_with_failure_during_playback_comparison(self, name, compare_in_dedicated_process):

        class Operation(object):
            def __init__(self, value=None, multiply_input=1):
                self._value = value
                self.multiply_input = multiply_input

            @property
            @self.tape_recorder.intercept_input('input')
            def input(self):
                return self._value

            @self.tape_recorder.operation()
            def execute(self):
                return self.input * self.multiply_input

        Operation(3).execute()
        Operation(4).execute()
        Operation(5).execute()

        playback_counter = [0]

        def playback_function(recording):
            playback_counter[0] += 1
            if playback_counter[0] == 2:
                raise Exception("playback error")
                operation = Operation(multiply_input=2)
            elif playback_counter[0] == 3:
                operation = Operation(multiply_input=100)
            else:
                operation = Operation()
            return operation.execute()

        def player(recording_id):
            return self.tape_recorder.play(recording_id, playback_function)

        start_date = datetime.utcnow() - timedelta(hours=1)
        end_date = datetime.utcnow() + timedelta(hours=1)
        playable_recordings = find_matching_recording_ids(
            self.tape_recorder,
            category=Operation.__name__,
            lookup_properties=RecordingLookupProperties(start_date=start_date, end_date=end_date),
        )

        runner = Equalizer(playable_recordings, player, result_extractor=return_value_result_extractor,
                           comparator=exact_comparator,
                           compare_execution_config=CompareExecutionConfig(
                               keep_results_in_comparison=False,
                               compare_in_dedicated_process=compare_in_dedicated_process
                           ))

        comparison = list(runner.run_comparison())

        self.assertEqual(EqualityStatus.Equal, comparison[0].comparator_status.equality_status)
        self.assertEqual(EqualityStatus.EqualizerFailure, comparison[1].comparator_status.equality_status)
        self.assertEqual('playback error', comparison[1].comparator_status.message)
        self.assertIsNotNone(comparison[1].recording_id)
        self.assertIsNone(comparison[1].playback)
        self.assertEqual(EqualityStatus.Failed, comparison[2].comparator_status.equality_status)

        self.assertIsNone(comparison[0].expected)
        self.assertIsNone(comparison[0].actual)

    @parameterized.expand([("compare_in_dedicated_process", True),
                           ("playback_same_process", False),
                           ])
    def test_with_failure_during_play_and_compare_method_other_process(self, name, compare_in_dedicated_process):

        class Operation(object):
            def __init__(self, value=None, multiply_input=1):
                self._value = value
                self.multiply_input = multiply_input

            @property
            @self.tape_recorder.intercept_input('input')
            def input(self):
                return self._value

            @self.tape_recorder.operation()
            def execute(self):
                return self.input * self.multiply_input

        Operation(3).execute()

        def playback_function(recording):
            operation = Operation()
            return operation.execute()

        def player(recording_id):
            return self.tape_recorder.play(recording_id, playback_function)

        start_date = datetime.utcnow() - timedelta(hours=1)
        end_date = datetime.utcnow() + timedelta(hours=1)
        playable_recordings = find_matching_recording_ids(
            self.tape_recorder,
            category=Operation.__name__,
            lookup_properties=RecordingLookupProperties(start_date=start_date, end_date=end_date),
        )

        def side_affect(*args, **kwargs):
            raise Exception('total failure')

        with patch('playback.studio.equalizer.Equalizer._play_and_compare_recording',
                   side_affect):
            runner = Equalizer(playable_recordings, player, result_extractor=return_value_result_extractor,
                               comparator=exact_comparator,
                               compare_execution_config=CompareExecutionConfig(
                                   keep_results_in_comparison=False,
                                   compare_in_dedicated_process=compare_in_dedicated_process
                               ))
            comparison = list(runner.run_comparison())

        self.assertEqual(EqualityStatus.EqualizerFailure, comparison[0].comparator_status.equality_status)
        self.assertEqual('total failure', comparison[0].comparator_status.message)
        self.assertIsNotNone(comparison[0].recording_id)
        self.assertIsNone(comparison[0].playback)
        self.assertIsNone(comparison[0].expected)
        self.assertIsNone(comparison[0].actual)

    @parameterized.expand([("compare_in_dedicated_process", True),
                           ("playback_same_process", False),
                           ])
    def test_comparison_with_meta_and_limit_filtering(self, name, compare_in_dedicated_process):

        class Operation(object):
            def __init__(self, value=None):
                self._value = value

            @property
            @self.tape_recorder.intercept_input('input')
            def input(self):
                return self._value

            @self.tape_recorder.operation(metadata_extractor=lambda op_self, meta=None: {'meta': meta})
            def execute(self, meta=None):
                return self.input

        Operation(3).execute('a')
        Operation(4).execute('b')
        Operation(5).execute('b')

        def playback_function(recording):
            return Operation().execute()

        def player(recording_id):
            return self.tape_recorder.play(recording_id, playback_function)

        with patch.object(InMemoryTapeCassette, 'iter_recording_ids',
                          wraps=self.tape_cassette.iter_recording_ids) as mock:
            start_date = datetime.utcnow() - timedelta(hours=1)
            playable_recordings = find_matching_recording_ids(
                self.tape_recorder,
                category=Operation.__name__,
                lookup_properties=RecordingLookupProperties(
                    start_date=start_date,
                    metadata={'meta': 'b'},
                    limit=1),
            )
            runner = Equalizer(playable_recordings, player, result_extractor=return_value_result_extractor,
                               comparator=exact_comparator,
                               compare_execution_config=CompareExecutionConfig(
                                   keep_results_in_comparison=True,
                                   compare_in_dedicated_process=compare_in_dedicated_process
                               ))

            comparison = list(runner.run_comparison())

        self.assertEqual(1, len(comparison))
        self.assertEqual(EqualityStatus.Equal, comparison[0].comparator_status.equality_status)

        mock.assert_called_with(Operation.__name__, start_date=start_date, end_date=None,
                                limit=1, metadata={'meta': 'b'})
        self.assertEqual(4, comparison[0].expected)
        self.assertEqual(4, comparison[0].actual)

    @parameterized.expand([("compare_in_dedicated_process", True),
                           ("playback_same_process", False),
                           ])
    def test_equal_comparison_with_message(self, name, compare_in_dedicated_process):

        class Operation(object):
            def __init__(self, value=None, multiply_input=1):
                self._value = value
                self.multiply_input = multiply_input

            @property
            @self.tape_recorder.intercept_input('input')
            def input(self):
                return self._value

            @self.tape_recorder.operation()
            def execute(self):
                return self.input * self.multiply_input

        Operation(3).execute()
        Operation(4).execute()
        Operation(5).execute()

        playback_counter = [0]

        def playback_function(recording):
            playback_counter[0] += 1
            if playback_counter[0] == 2:
                operation = Operation(multiply_input=2)
            elif playback_counter[0] == 3:
                operation = Operation(multiply_input=100)
            else:
                operation = Operation()
            return operation.execute()

        def player(recording_id):
            return self.tape_recorder.play(recording_id, playback_function)

        start_date = datetime.utcnow() - timedelta(hours=1)
        end_date = datetime.utcnow() + timedelta(hours=1)
        playable_recordings = find_matching_recording_ids(
            self.tape_recorder,
            category=Operation.__name__,
            lookup_properties=RecordingLookupProperties(
                start_date=start_date,
                end_date=end_date)
        )
        runner = Equalizer(playable_recordings, player, result_extractor=return_value_result_extractor,
                           comparator=exact_comparator_with_message,
                           compare_execution_config=CompareExecutionConfig(
                               keep_results_in_comparison=True,
                               compare_in_dedicated_process=compare_in_dedicated_process
                           ))

        comparison = list(runner.run_comparison())

        self.assertEqual(EqualityStatus.Equal, comparison[0].comparator_status.equality_status)
        self.assertEqual('{status} - {status}'.format(
            status=comparison[0].comparator_status.equality_status.name), str(comparison[0].comparator_status))
        self.assertEqual(EqualityStatus.Different, comparison[1].comparator_status.equality_status)
        self.assertEqual('{status} - {status}'.format(
            status=comparison[1].comparator_status.equality_status.name), str(comparison[1].comparator_status))
        self.assertEqual(EqualityStatus.Failed, comparison[2].comparator_status.equality_status)
        self.assertEqual('{status} - {status}'.format(
            status=comparison[2].comparator_status.equality_status.name), str(comparison[2].comparator_status))

        self.assertEqual(EqualityStatus.Equal.name, comparison[0].comparator_status.message)
        self.assertEqual(EqualityStatus.Different.name, comparison[1].comparator_status.message)
        self.assertEqual(EqualityStatus.Failed.name, comparison[2].comparator_status.message)

    @parameterized.expand([("compare_in_dedicated_process", True),
                           ("playback_same_process", False),
                           ])
    def test_equal_comparison_comparator_data_extraction(self, name, compare_in_dedicated_process):

        class Operation(object):
            def __init__(self, value=None, override_input=None):
                self._value = value
                self.override_input = override_input

            @property
            @self.tape_recorder.intercept_input('input')
            def input(self):
                return self._value

            @self.tape_recorder.operation()
            def execute(self):
                if self.override_input:
                    return self.override_input
                return self.input

        Operation(3).execute()
        Operation(4).execute()
        Operation(5).execute()

        playback_counter = [0]

        def playback_function(recording):
            playback_counter[0] += 1
            if playback_counter[0] == 2:
                operation = Operation(override_input=100)
            else:
                operation = Operation()
            return operation.execute()

        def player(recording_id):
            return self.tape_recorder.play(recording_id, playback_function)

        def comparison_data_extractor(recording):
            self.assertIsNotNone(recording)
            return {'multiplier': 1}

        def exact_comparator_with_multiplier(recorded_result, playback_result, multiplier):
            if playback_result >= 10:
                playback_result *= 0
                recorded_result *= 0
            else:
                playback_result *= multiplier
                recorded_result *= multiplier

            return ComparatorResult(
                EqualityStatus.Equal if recorded_result * multiplier == playback_result * multiplier
                else EqualityStatus.Different, str(multiplier))

        start_date = datetime.utcnow() - timedelta(hours=1)
        end_date = datetime.utcnow() + timedelta(hours=1)
        playable_recordings = find_matching_recording_ids(
            self.tape_recorder,
            category=Operation.__name__,
            lookup_properties=RecordingLookupProperties(
                start_date=start_date,
                end_date=end_date)
        )
        runner = Equalizer(playable_recordings, player, result_extractor=return_value_result_extractor,
                           comparator=exact_comparator_with_multiplier,
                           comparison_data_extractor=comparison_data_extractor,
                           compare_execution_config=CompareExecutionConfig(
                               keep_results_in_comparison=True,
                               compare_in_dedicated_process=compare_in_dedicated_process
                           ))

        comparison = list(runner.run_comparison())

        self.assertEqual(EqualityStatus.Equal, comparison[0].comparator_status.equality_status)
        self.assertEqual(EqualityStatus.Equal, comparison[1].comparator_status.equality_status)
        self.assertEqual(EqualityStatus.Equal, comparison[2].comparator_status.equality_status)

        self.assertEqual('1', comparison[0].comparator_status.message)
        self.assertEqual('1', comparison[1].comparator_status.message)
        self.assertEqual('1', comparison[2].comparator_status.message)

    def test_random_sample(self):

        class Operation(object):
            def __init__(self, value=None):
                self._value = value

            @property
            @self.tape_recorder.intercept_input('input')
            def input(self):
                return self._value

            @self.tape_recorder.operation()
            def execute(self):
                return self.input

        for i in range(10):
            Operation(i).execute()

        def playback_function(recording):
            return Operation().execute()

        random.seed(12)
        first_list = None
        start_date = datetime.utcnow() - timedelta(hours=1)
        playable_recordings = find_matching_recording_ids(
            self.tape_recorder,
            category=Operation.__name__,
            lookup_properties=RecordingLookupProperties(
                start_date=start_date,
                random_sample=True,
                limit=3)
        )
        first_list = list(playable_recordings)
        self.assertEqual(len(first_list), 3)

        random.seed(4)
        start_date = datetime.utcnow() - timedelta(hours=1)
        playable_recordings = find_matching_recording_ids(
            self.tape_recorder,
            category=Operation.__name__,
            lookup_properties=RecordingLookupProperties(
                start_date=start_date,
                random_sample=True,
                limit=3)
        )
        second_list = list(playable_recordings)
        self.assertNotEqual(second_list, first_list)

    @parameterized.expand([("compare_in_dedicated_process", True),
                           ("playback_same_process", False),
                           ])
    def test_run_with_specific_ids(self, name, compare_in_dedicated_process):

        class Operation(object):
            def __init__(self, value=None, override_input=None):
                self._value = value
                self.override_input = override_input

            @property
            @self.tape_recorder.intercept_input('input')
            def input(self):
                return self._value

            @self.tape_recorder.operation()
            def execute(self):
                if self.override_input:
                    return self.override_input
                return self.input

        Operation(3).execute()
        id1 = self.tape_cassette.get_last_recording_id()
        Operation(4).execute()
        Operation(5).execute()
        id2 = self.tape_cassette.get_last_recording_id()
        Operation(6).execute()

        playback_counter = [0]

        def playback_function(recording):
            playback_counter[0] += 1
            if playback_counter[0] == 2:
                operation = Operation(override_input=100)
            else:
                operation = Operation()
            return operation.execute()

        def player(recording_id):
            return self.tape_recorder.play(recording_id, playback_function)

        runner = Equalizer([id1, id2], player, result_extractor=return_value_result_extractor,
                           comparator=exact_comparator,
                           compare_execution_config=CompareExecutionConfig(
                               keep_results_in_comparison=True,
                               compare_in_dedicated_process=compare_in_dedicated_process
                           ))

        comparison = list(runner.run_comparison())

        self.assertEqual(id1, comparison[0].playback.original_recording.id)
        self.assertEqual(id2, comparison[1].playback.original_recording.id)
        self.assertEqual(2, len(comparison))

    def test_equalizer_recycle_process_thread(self):
        class Operation(object):
            def __init__(self, value=None):
                self._value = value

            @property
            @self.tape_recorder.intercept_input('input')
            def input(self):
                return self._value

            @self.tape_recorder.operation()
            def execute(self):
                return self.input

        for i in range(20):
            Operation(i).execute()

        def playback_function(recording):
            return Operation().execute()

        def player(recording_id):
            return self.tape_recorder.play(recording_id, playback_function)

        start_date = datetime.utcnow() - timedelta(hours=1)
        end_date = datetime.utcnow() + timedelta(hours=1)
        playable_recordings = find_matching_recording_ids(
            self.tape_recorder,
            category=Operation.__name__,
            lookup_properties=RecordingLookupProperties(start_date=start_date, end_date=end_date),
        )

        runner = Equalizer(playable_recordings, player, result_extractor=return_value_result_extractor,
                           comparator=exact_comparator,
                           compare_execution_config=CompareExecutionConfig(
                               keep_results_in_comparison=True,
                               compare_in_dedicated_process=True,
                               compare_process_recycle_rate=2
                           ))

        with patch.object(Equalizer, '_create_new_player_process',
                          wraps=runner._create_new_player_process) as wrapped:
            comparison = list(runner.run_comparison())

        self.assertEqual(10, wrapped.call_count)

        self.assertEqual(20, len(comparison))
        for c in comparison:
            self.assertEqual(EqualityStatus.Equal, c.comparator_status.equality_status)

    def test_equalizer_process_timeout(self):
        class Operation(object):
            def __init__(self, value=None, playback=False):
                self._value = value
                self._playback = playback

            @property
            @self.tape_recorder.intercept_input('input')
            def input(self):
                return self._value

            @self.tape_recorder.operation()
            def execute(self):
                res = self.input
                if res == 2 and self._playback:
                    sleep(30)
                return res

        for i in range(10):
            Operation(i).execute()

        def playback_function(recording):
            return Operation(playback=True).execute()

        def player(recording_id):
            return self.tape_recorder.play(recording_id, playback_function)

        start_date = datetime.utcnow() - timedelta(hours=1)
        end_date = datetime.utcnow() + timedelta(hours=1)
        playable_recordings = find_matching_recording_ids(
            self.tape_recorder,
            category=Operation.__name__,
            lookup_properties=RecordingLookupProperties(start_date=start_date, end_date=end_date),
        )

        runner = Equalizer(playable_recordings, player, result_extractor=return_value_result_extractor,
                           comparator=exact_comparator,
                           compare_execution_config=CompareExecutionConfig(
                               keep_results_in_comparison=True,
                               compare_in_dedicated_process=True,
                               compare_process_timeout=1,
                           ))

        original_kill = runner._kill_compare_process

        def side_affect(self):
            original_kill()
            # To have code coverage of OSError being raised
            raise OSError('bla')

        start_time = time()
        with patch('playback.studio.equalizer.Equalizer._kill_compare_process',
                   side_affect):
            comparison = list(runner.run_comparison())
        duration = time() - start_time

        self.assertLessEqual(duration, 5)
        self.assertEqual(EqualityStatus.EqualizerFailure, comparison[2].comparator_status.equality_status)
        self.assertIn('timeout', comparison[2].comparator_status.message.lower())
        del comparison[2]
        for c in comparison:
            self.assertEqual(EqualityStatus.Equal, c.comparator_status.equality_status)

    def test_equalizer_process_died(self):
        class Operation(object):
            def __init__(self, value=None, playback=False):
                self._value = value
                self._playback = playback

            @property
            @self.tape_recorder.intercept_input('input')
            def input(self):
                return self._value

            @self.tape_recorder.operation()
            def execute(self):
                res = self.input
                if res == 2 and self._playback:
                    exit(1)
                return res

        for i in range(10):
            Operation(i).execute()

        def playback_function(recording):
            return Operation(playback=True).execute()

        def player(recording_id):
            return self.tape_recorder.play(recording_id, playback_function)

        start_date = datetime.utcnow() - timedelta(hours=1)
        end_date = datetime.utcnow() + timedelta(hours=1)
        playable_recordings = find_matching_recording_ids(
            self.tape_recorder,
            category=Operation.__name__,
            lookup_properties=RecordingLookupProperties(start_date=start_date, end_date=end_date),
        )

        runner = Equalizer(playable_recordings, player, result_extractor=return_value_result_extractor,
                           comparator=exact_comparator,
                           compare_execution_config=CompareExecutionConfig(
                               keep_results_in_comparison=True,
                               compare_in_dedicated_process=True,
                               compare_process_timeout=1,
                           ))

        start_time = time()
        comparison = list(runner.run_comparison())
        duration = time() - start_time

        self.assertLessEqual(duration, 5)
        self.assertEqual(EqualityStatus.EqualizerFailure, comparison[2].comparator_status.equality_status)
        self.assertIn('died', comparison[2].comparator_status.message.lower())
        del comparison[2]
        for c in comparison:
            self.assertEqual(EqualityStatus.Equal, c.comparator_status.equality_status)



# p3ready
from __future__ import absolute_import

import shutil
import unittest
from datetime import datetime, timedelta
from tempfile import mkdtemp

import six

from playback.exceptions import NoSuchRecording
from playback.tape_cassettes.file_based.file_based_tape_cassette import FileBasedTapeCassette


def assert_items_equal(testcase, seq1, seq2, msg=""):
    if six.PY3:
        testcase.assertCountEqual(seq1, seq2, msg)
    else:
        testcase.assertItemsEqual(seq1, seq2, msg)


class TestFileBasedTapeCassette(unittest.TestCase):

    def setUp(self):
        self.directory = mkdtemp()
        # Delete dir so the cassette will create it
        shutil.rmtree(self.directory)
        self.cassette = FileBasedTapeCassette(self.directory)

    def tearDown(self):
        self.cassette.close()
        shutil.rmtree(self.directory)

    def test_create_save_and_fetch_empty_recording(self):
        recording = self.cassette.create_new_recording('test_operation')
        self.cassette.save_recording(recording)
        fetched_recording = self.cassette.get_recording(recording.id)
        self.assertEqual(recording.id, fetched_recording.id)

    def test_get_no_such_recording(self):
        with self.assertRaises(NoSuchRecording):
            self.cassette.get_recording('nonexisting')

    def test_create_save_and_fetch_recording_with_data(self):
        recording = self.cassette.create_new_recording('test_operation')
        recording.set_data('key1', 5)
        recording.set_data('key2', {'obj_key1': 2, 'obj_key2': "bla"})
        self.cassette.save_recording(recording)
        fetched_recording = self.cassette.get_recording(recording.id)
        self.assertEqual(recording.id, fetched_recording.id)

        self.assertEqual(5, recording.get_data('key1'))
        self.assertEqual(5, recording['key1'])
        self.assertEqual(recording.get_data('key1'), fetched_recording.get_data('key1'))

        self.assertEqual({'obj_key1': 2, 'obj_key2': "bla"}, recording.get_data('key2'))
        self.assertEqual(recording.get_data('key2'), fetched_recording.get_data('key2'))

        assert_items_equal(self, ['key1', 'key2'], recording.get_all_keys())
        assert_items_equal(self, recording.get_all_keys(), fetched_recording.get_all_keys())

    def test_create_save_and_fetch_recording_with_metadata(self):
        recording = self.cassette.create_new_recording('test_operation')
        metadata = {'key1': 5, 'key2': {'obj_key1': 2, 'obj_key2': "bla"}}
        recording.add_metadata(metadata)
        self.cassette.save_recording(recording)
        fetched_recording = self.cassette.get_recording(recording.id)
        self.assertEqual(recording.id, fetched_recording.id)

        self.assertEqual(metadata, recording.get_metadata())
        self.assertEqual(recording.get_metadata(), fetched_recording.get_metadata())

    def test_fetch_recording_ids_by_category(self):
        recording1 = self.cassette.create_new_recording('test_operation1')
        self.cassette.save_recording(recording1)
        recording2 = self.cassette.create_new_recording('test_operation1')
        self.cassette.save_recording(recording2)
        recording3 = self.cassette.create_new_recording('test_operation2')
        self.cassette.save_recording(recording3)

        assert_items_equal(self, [recording1.id, recording2.id],
                           list(self.cassette.iter_recording_ids(category='test_operation1')))
        assert_items_equal(self, [recording3.id],
                           list(self.cassette.iter_recording_ids(category='test_operation2')))

    def test_fetch_recording_ids_by_category_date_and_metadata(self):
        recording1 = self.cassette.create_new_recording('test_operation1')
        recording1.add_metadata({'property': True})
        self.cassette.save_recording(recording1)
        recording2 = self.cassette.create_new_recording('test_operation1')
        recording2.add_metadata({'property': False})
        self.cassette.save_recording(recording2)
        recording3 = self.cassette.create_new_recording('test_operation2')
        self.cassette.save_recording(recording3)

        assert_items_equal(self, [recording2.id],
                           list(self.cassette.iter_recording_ids(category='test_operation1',
                                                                 start_date=datetime.utcnow() - timedelta(hours=1),
                                                                 end_date=datetime.utcnow() + timedelta(hours=1),
                                                                 metadata={'property': False})))

    def test_fetch_recordings_metadata_by_category_date_and_metadata(self):
        recording1 = self.cassette.create_new_recording('test_operation1')
        recording1.add_metadata({'property': True})
        self.cassette.save_recording(recording1)
        recording2 = self.cassette.create_new_recording('test_operation1')
        recording2.add_metadata({'property': False})
        self.cassette.save_recording(recording2)
        recording3 = self.cassette.create_new_recording('test_operation2')
        self.cassette.save_recording(recording3)

        assert_items_equal(self, [{'property': False}],
                           list(self.cassette.iter_recordings_metadata(
                               category='test_operation1',
                               start_date=datetime.utcnow() - timedelta(hours=1),
                               end_date=datetime.utcnow() + timedelta(hours=1),
                               metadata={'property': False})))

    def test_fetch_recording_ids_by_category_and_limit(self):
        recording1 = self.cassette.create_new_recording('test_operation1')
        self.cassette.save_recording(recording1)
        recording2 = self.cassette.create_new_recording('test_operation1')
        self.cassette.save_recording(recording2)
        recording3 = self.cassette.create_new_recording('test_operation1')
        self.cassette.save_recording(recording3)

        self.assertEqual(2, len(list(self.cassette.iter_recording_ids(category='test_operation1', limit=2))))

    def test_extract_recording_category(self):
        recording1 = self.cassette.create_new_recording('test_operation1')
        self.assertEqual('test_operation1', self.cassette.extract_recording_category(recording1.id))
        recording2 = self.cassette.create_new_recording('test_operation2')
        self.assertEqual('test_operation2', self.cassette.extract_recording_category(recording2.id))



# We create the tape recorder singleton that will be used by the endpoints
import os

from playback.tape_cassettes.asynchronous.async_record_only_tape_cassette import AsyncRecordOnlyTapeCassette
from playback.tape_cassettes.file_based.file_based_tape_cassette import FileBasedTapeCassette
from playback.tape_recorder import TapeRecorder

tape_recorder = TapeRecorder(None)
tape_recorder.disable_recording()

# For demonstration purpose, persist file in underline recordings dir
recordings_path = os.path.dirname(os.path.realpath(__file__)) + '/recordings'


def init_recording_mode():
    """
    Initialize and start the tape cassette and tape recorder for recording mode.
    in order for the recording not to be blocking and part of the operation latency, we wrap it with the asynchronous
    tape cassette which will flush the recording to a file in a separate thread, reducing the impact on the thread
    that executes the operation
    """
    print("Starting recording mode - recordings will be saved to {}".format(recordings_path))
    tape_cassette = AsyncRecordOnlyTapeCassette(FileBasedTapeCassette(recordings_path))
    tape_cassette.start()
    tape_recorder.tape_cassette = tape_cassette
    # This will activate the tape recorder in recording mode
    tape_recorder.enable_recording()


def init_playback_mode():
    print("Starting playback mode - recordings will be read from {}".format(recordings_path))
    tape_cassette = FileBasedTapeCassette(recordings_path)
    tape_recorder.tape_cassette = tape_cassette
